//
//  float16.h
//  16bit
//
//  Created by 管胡昊 on 2020/2/5.
//  Copyright © 2020 管胡昊. All rights reserved.
//

#ifndef FLOAT16_H
#define FLOAT16_H

struct float16
{
    //private member variable
private:
    /* 
    sign is the sign bit 1 means negative, 0 means positive
    exp is the exponent with 16 offset
    data is the data,similar to ieee-754,the highest is default 1 and ignored 
    */


    // mask for calculate the highest 1
    static unsigned int mask[32];
    static unsigned int pow2[32];
    
    // private function
    int FindHighOne(const int &num, int &l, int &r);
    int AbsCompare(const float16 & a,const float16 & b);

public:
    unsigned short data : 10;
    unsigned short exp : 5;
    unsigned short sign : 1;

    float16 SetOverFlow();

    // judge whether overflow
    int IsOverlFlow() const;
    
    /* constructor by sign, exp, data
       sign:1bit exp:5bit data:10bit similar to ieee 32 floating point */
    float16(const int& s, const int& e, const int& d);

    /* default constructor
       This initializes the 16bit floating point to 0. */
    float16();

    // constructor by a 32-bit floating point
    float16(const float& data);
    template<class T> float16(const T& data);

    // constructor by other datatype
    //template<class T> float16(const T &data);

    // change float16 to flaot as you can see the result is a 32-bit floating point
    float Float();
    
    /* assignment function and tempalte function
       float assignment function is the basic function
       template assignment function is force change other datetype to float 
       then call the float assignment function
       template assignment function now support int, double */
    float16 operator = (const float16& data);
    float16 operator = (const float& data);
    template<class T> float16 operator = (const T& data);

    // overload operator (less than) eg. a<b
    int operator < (const float16& data);
    template<class T> int operator <(const T& data);

    // overload opertator <= (less or equal than) a<=b
    int operator <= (const float16& data);
    template<class T> int operator <=(const T& data);

    // overload operator (greater than) eg. a>b
    int operator > (const float16& data);
    template<class T> int operator >(const T& data);

    //overload opertator <= (greater or equal than) a>=b
    int operator >= (const float16& data);
    template<class T> int operator >=(const T& data);

    // overload operator + (add) eg. a+b
    float16 operator + (const float16& data);
    template<class T> float16 operator +(const T& data);

    // overload operator += (add) eg. a+=b
    float16 operator += (const float16& data);
    template<class T> float16 operator +=(const T& data);

    // overload operator -(negetive) eg. -a
    float16 operator - ();

    // overload operator - (substraction) eg. a-b
    float16 operator - (const float16& data);
    template<class T> float16 operator -(const T& data);

    // overload operator -= (substraction) eg. a-=b
    float16 operator -= (const float16& data);
    template<class T> float16 operator -=(const T& data);

    // overload operator * (multiple) eg. a*b
    float16 operator * (const float16& data);
    template<class T> float16 operator *(const T& data);

    // overload operator *= (multiple) eg. a*=b
    float16 operator *= (const float16& data);
    template<class T> float16 operator *=(const T& data);

    // overload operator / (division) eg. a/b
    float16 GetInverse() const;
    float16 operator / (const float16& data);
    template<class T> float16 operator /(const T& data);

    // overload operator /= (division) eg. a/=b
    float16 operator /= (const float16& data);
    template<class T> float16 operator /=(const T& data);

};
#endif /* FLOAT16_H */
